import typing

import pytest
from matplotlib.figure import Figure
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from slugify import slugify

import outset
from outset._auxlib.resize_figure_to_axes_ import resize_figure_to_axes


@pytest.mark.parametrize(
    "plot", [sns.scatterplot, outset.marqueeplot, lambda *args, **kwargs: None]
)
def test_resize_figure_to_axes(plot: typing.Callable):
    plt.close("all")
    # Setup a figure and axes
    fig, (ax1, ax2) = plt.subplots(1, 2)
    plot(
        data=pd.DataFrame({"x": [0, 1, 2], "y": [0, 1, 2]}),
        ax=ax1,
        x="x",
        y="y",
    )

    outpath = f"/tmp/test_resize_figure_to_axes_{slugify(plot.__name__)}_a.png"
    plt.savefig(outpath)
    print(f"saved graphic to {outpath}")

    ax2.set_position([0.2, 0.2, 0.2, 0.2])

    outpath = f"/tmp/test_resize_figure_to_axes_{slugify(plot.__name__)}_b.png"
    plt.savefig(outpath)
    print(f"saved graphic to {outpath}")

    # Save the original size and position
    before_size = fig.get_size_inches()
    before_pos_bounds = ax1.get_position().bounds
    before_pos_in_inches = [
        before_pos_bounds[0] * before_size[0],
        before_pos_bounds[1] * before_size[1],
        before_pos_bounds[2] * before_size[0],
        before_pos_bounds[3] * before_size[1],
    ]

    # Call the function
    resize_figure_to_axes(fig, ax1)

    # Get the new size and position
    after_size = fig.get_size_inches()
    after_pos_bounds = ax1.get_position().bounds
    after_pos_in_inches = [
        after_pos_bounds[0] * after_size[0],
        after_pos_bounds[1] * after_size[1],
        after_pos_bounds[2] * after_size[0],
        after_pos_bounds[3] * after_size[1],
    ]

    outpath = f"/tmp/test_resize_figure_to_axes_{slugify(plot.__name__)}_c.png"
    plt.savefig(outpath)
    print(f"saved graphic to {outpath}")

    # Check that the size has changed
    assert tuple(after_size) != tuple(before_size)

    # Check that axes position in inches is approximately the same
    assert np.allclose(after_pos_in_inches, before_pos_in_inches, atol=0.01)


@pytest.mark.parametrize(
    "plot", [sns.scatterplot, lambda *args, **kwargs: None]
)
def test_resize_figure_to_axes_FacetGrid(plot: typing.Callable):
    plt.close("all")
    # Setup a figure and axes
    g = sns.FacetGrid(
        pd.DataFrame({"x": [0, 1, 2], "y": [0, 1, 2], "col": [0, 1, 1]}),
        col="col",
    )
    g.map(plot, "x", "y")

    g.axes.flat[1].set_position([0.2, 0.2, 0.2, 0.2])

    # Save the original size and position
    before_size = g.figure.get_size_inches()
    before_pos_bounds = g.axes.flat[0].get_position().bounds
    before_pos_in_inches = [
        before_pos_bounds[0] * before_size[0],
        before_pos_bounds[1] * before_size[1],
        before_pos_bounds[2] * before_size[0],
        before_pos_bounds[3] * before_size[1],
    ]

    # Call the function
    resize_figure_to_axes(g.figure, g.axes.flat[0])

    # Get the new size and position
    after_size = g.figure.get_size_inches()
    after_pos_bounds = g.axes.flat[0].get_position().bounds
    after_pos_in_inches = [
        after_pos_bounds[0] * after_size[0],
        after_pos_bounds[1] * after_size[1],
        after_pos_bounds[2] * after_size[0],
        after_pos_bounds[3] * after_size[1],
    ]

    outpath = (
        f"/tmp/test_resize_figure_to_axes_facet_{slugify(plot.__name__)}.png"
    )
    plt.savefig(outpath)
    print(f"saved graphic to {outpath}")

    # Check that the size has changed
    assert tuple(after_size) != tuple(before_size)

    # Check that axes position in inches is approximately the same
    assert np.allclose(after_pos_in_inches, before_pos_in_inches, atol=0.01)


@pytest.mark.parametrize(
    "plot", [sns.scatterplot, lambda *args, **kwargs: None]
)
def test_resize_figure_to_axes_OutsetGrid(plot: typing.Callable):
    plt.close("all")
    # Setup a figure and axes
    g = outset.OutsetGrid(
        pd.DataFrame({"x": [0, 1, 2], "y": [0, 1, 2], "col": [1, 1, 1]}),
        col="col",
        x="x",
        y="y",
    )
    g.map_dataframe(plot, x="x", y="y", legend=False)

    outpath = (
        f"/tmp/test_resize_figure_to_axes_outset_{slugify(plot.__name__)}_a.png"
    )
    plt.savefig(outpath)
    print(f"saved graphic to {outpath}")

    g.axes.flat[1].set_position([0.2, 0.2, 0.2, 0.2])

    # Save the original size and position
    outpath = (
        f"/tmp/test_resize_figure_to_axes_outset_{slugify(plot.__name__)}_b.png"
    )
    plt.savefig(outpath)
    print(f"saved graphic to {outpath}")

    before_size = g.figure.get_size_inches()
    before_pos_bounds = g.axes.flat[0].get_position().bounds
    before_pos_in_inches = [
        before_pos_bounds[0] * before_size[0],
        before_pos_bounds[1] * before_size[1],
        before_pos_bounds[2] * before_size[0],
        before_pos_bounds[3] * before_size[1],
    ]

    # Call the function
    resize_figure_to_axes(g.figure, g.axes.flat[0])

    # Get the new size and position
    after_size = g.figure.get_size_inches()
    after_pos_bounds = g.axes.flat[0].get_position().bounds
    after_pos_in_inches = [
        after_pos_bounds[0] * after_size[0],
        after_pos_bounds[1] * after_size[1],
        after_pos_bounds[2] * after_size[0],
        after_pos_bounds[3] * after_size[1],
    ]

    g.marqueeplot()

    outpath = (
        f"/tmp/test_resize_figure_to_axes_outset_{slugify(plot.__name__)}_c.png"
    )
    plt.savefig(outpath)
    print(f"saved graphic to {outpath}")

    # Check that the size has changed
    assert tuple(after_size) != tuple(before_size)

    # # Check that axes position in inches is approximately the same
    assert np.allclose(after_pos_in_inches, before_pos_in_inches, atol=0.01)



"""Top-level package for outset."""

__version__ = "0.0.0""-rc183-post1"

from ._draw_marquee import draw_marquee
from ._inset_outsets import inset_outsets
from ._marqueeplot import marqueeplot
from ._OutsetGrid import OutsetGrid

__all__ = [
    "draw_marquee",
    "inset_outsets",
    "marqueeplot",
    "OutsetGrid",
]



from outset._auxlib.minimal_perimeter_permutation_ import (
    minimal_perimeter_permutation,
)


def test_basic_functionality():
    # Test the basic functionality with a simple example
    points = [(0, 0), (1, 1), (2, 2)]
    expected = [(0, 0), (1, 1), (2, 2)]
    assert minimal_perimeter_permutation(points) == expected


def test_single_point():
    # Test the function with a single point
    points = [(1, 1)]
    expected = [(1, 1)]
    assert minimal_perimeter_permutation(points) == expected


def test_empty_input():
    # Test the function with an empty list
    points = []
    expected = []
    assert minimal_perimeter_permutation(points) == expected


def test_non_trivial_case():
    # Test the function with a non-trivial case
    points = [(0, 0), (2, 2), (1, 1), (3, 3)]
    # Expected output should be in the order that minimizes the perimeter
    # The exact expected output might depend on the implementation
    result = minimal_perimeter_permutation(points)
    assert len(result) == len(points)  # Check if all points are included
    # Further checks can be added based on expected behavior


def test_negative_coordinates():
    # Test with negative coordinates
    points = [(-1, -1), (0, 0), (1, 1)]
    expected = [(-1, -1), (0, 0), (1, 1)]
    assert minimal_perimeter_permutation(points) == expected


def test_duplicate_points():
    # Test with duplicate points
    points = [(0, 0), (1, 1), (1, 1)]
    expected = [(0, 0), (1, 1), (1, 1)]
    assert minimal_perimeter_permutation(points) == expected



import typing
import matplotlib.pyplot as plt
from matplotlib import axes as mpl_axes
import numpy as np


class TweakSpreadArea:
    """Functor to spread apart a crowded axes area.

    The adjustment is applied by moving the vertex away from the centroid of the parametrized x/y limits, scaled by a spread factor.
    """

    _spread_factor: typing.Tuple[float, float]
    _xlim: typing.Optional[typing.Tuple[float, float]]
    _ylim: typing.Optional[typing.Tuple[float, float]]

    def __init__(
        self: "TweakSpreadArea",
        spread_factor: typing.Union[float, typing.Tuple[float, float]] = 2.0,
        *,
        xlim: typing.Optional[typing.Tuple[float, float]] = None,
        ylim: typing.Optional[typing.Tuple[float, float]] = None,
    ) -> None:
        """
        Initialize the TweakSpreadArea functor.

        Parameters
        ----------
        spread_factor : Union[float, Tuple[float, float]], optional
            How aggressively to spread apart x and/or y coordinates.

            If a tuple is provided, the first element is used for x-coordinate
            spread and the second for y-coordinate spread. If a single float is
            provided, it is used for both.
        xlim : Optional[Tuple[float, float]], optional
            If provided, outer vertices falling within the x-coordinate range will be spread horizontally away from the range's centroid.
        ylim : Optional[Tuple[float, float]], optional
            If provided, outer vertices falling within the y-coordinate range will be spread vertically away from the range's centroid.
        """

        if isinstance(spread_factor, tuple):
            self._spread_factor = spread_factor
        else:
            self._spread_factor = (spread_factor, spread_factor)
        self._xlim = xlim
        self._ylim = ylim

    def __call__(
        self: "TweakSpreadArea",
        leader_vertices: typing.Sequence[typing.Tuple[float, float]],
        ax: typing.Optional[mpl_axes.Axes] = None,
    ) -> typing.List[typing.Tuple[float, float]]:
        """Spread outer vertex of leader polygon out from specied area.

        Parameters
        ----------
        leader_vertices : Sequence[Tuple[float, float]]
            A sequence of tuples representing the `(x, y)` coordinates of the
            leader vertices.
        ax : mpl_axes.Axes
            The matplotlib Axes object on which the leader vertices are plotted.

            Not used for this tweak.
        Returns
        -------
        List[Tuple[float, float]]
            The modified sequence of leader vertices with the outer vertex
            position adjusted.
        """
        x, y = leader_vertices[-1]

        if self._xlim is not None and np.clip(x, *self._xlim) == x:
            x_centroid = np.mean(self._xlim)
            x = x_centroid + (x - x_centroid) * self._spread_factor[0]

        if self._ylim is not None and np.clip(y, *self._ylim) == y:
            y_centroid = np.mean(self._ylim)
            y = y_centroid + (y - y_centroid) * self._spread_factor[1]

        return [*leader_vertices[:-1], (x, y)]



"""Create margin annotations for data outside axes viewport."""

from ._CalcBoundsIQR import CalcBoundsIQR
from ._rescale_clip_outliers import rescale_clip_outliers
from ._stub_all_clipped_values import stub_all_clipped_values
from ._stub_edge_mark import stub_edge_mark

__all__ = [
    "CalcBoundsIQR",
    "rescale_clip_outliers",
    "stub_all_clipped_values",
    "stub_edge_mark",
]



import colorsys
import typing

import matplotlib.colors as mpl_colors


# adapted from https://stackoverflow.com/a/49601444
def lighten_color(
    color: typing.Union[str, typing.Tuple], amount: float = 0.5
) -> typing.Tuple:
    """Lightens the given color by multiplying (1-luminosity) by the given amount. Input can be matplotlib color string, hex string, or RGB tuple.

    Examples:
    >> lighten_color('g', 0.3)
    >> lighten_color('#F034A3', 0.6)
    >> lighten_color((.3,.55,.1), 0.5)
    """

    try:
        c = mpl_colors.cnames[color]
    except:
        c = color
    c = colorsys.rgb_to_hls(*mpl_colors.to_rgb(c))
    return colorsys.hls_to_rgb(c[0], 1 - amount * (1 - c[1]), c[2])



from matplotlib import pyplot as plt
import pandas as pd
import pytest

import outset


@pytest.mark.integration
def test_make_taxonomy():
    # Create sample data with a 'outset' column for grouping
    data = pd.DataFrame(
        {
            "x": [0, 2, 5, 6],
            "y": [1, 4, 3, 9],
            "outset": [
                "group1",
                "group1",
                "group2",
                "group2",
            ],
        }
    )

    g = outset.OutsetGrid(data=data, x="x", y="y", col="outset", hue="outset")
    g.marqueeplot()

    g.source_axes.annotate(
        "callout mark",
        xy=(3, 5),
        xytext=(3.5, 4),
        horizontalalignment="left",
        arrowprops=dict(arrowstyle="->", lw=1),
    )
    g.source_axes.annotate(
        "callout leader",
        xy=(2.5, 4),
        xytext=(3.5, 3),
        horizontalalignment="left",
        arrowprops=dict(arrowstyle="->", lw=1),
    )
    g.source_axes.annotate(
        "frame",
        xy=(1, 2),
        xytext=(1, 2),
        horizontalalignment="center",
    )
    g.source_axes.annotate(
        "marquee",
        xy=(1.5, 6),
        xytext=(1.5, 6.5),
        ha="center",
        va="bottom",
        arrowprops=dict(
            arrowstyle="-[, widthB=4.0, lengthB=1.0", lw=2.0, color="k"
        ),
    )
    g.broadcast_source(lambda ax: ax.set_title("source axes"))
    g.broadcast_outset(lambda ax: ax.set_title("outset axes"))

    outpath = "/tmp/test_taxonomy_outset.png"
    plt.savefig(outpath)
    print(f"saved graphic to {outpath}")

    outset.inset_outsets(g, insets="NW")
    g.source_axes.annotate(
        "inset outset axes",
        xy=(1.5, 10.9),
        xytext=(1.5, 11),
        ha="center",
        va="bottom",
        arrowprops=dict(
            arrowstyle="-[, widthB=5, lengthB=0.5", lw=2.0, color="k"
        ),
    )
    outset.inset_outsets(g, insets="NW")
    g.source_axes.annotate(
        "",
        xy=(2.5, 11.9),
        xytext=(2.5, 12),
        ha="center",
        va="bottom",
        arrowprops=dict(
            arrowstyle="-[, widthB=10, lengthB=0.5", lw=2.0, color="k"
        ),
        annotation_clip=False,
    )

    # additional annotations for other marquees
    # g.source_axes.annotate(
    #     "marquee",
    #     xy=(1.5, 10.5),
    #     xytext=(4.5, 11.5),
    #     ha="center",
    #     arrowprops=dict(arrowstyle="->", lw=1),
    # )
    # g.source_axes.annotate(
    #     "",
    #     xy=(3.25, 10.75),
    #     xytext=(4, 11.25),
    #     ha="center",
    #     arrowprops=dict(arrowstyle="->", lw=1),
    # )
    # g.source_axes.annotate(
    #     "",
    #     xy=(5.5, 8),
    #     xytext=(4.5, 11.25),
    #     ha="center",
    #     arrowprops=dict(arrowstyle="->", lw=1),
    # )

    outpath = "/tmp/test_taxonomy_inset.png"
    plt.savefig(outpath)
    print(f"saved graphic to {outpath}")



import itertools as it
import typing

import numpy as np


def minimal_perimeter_permutation(
    points: typing.Iterable[typing.Tuple[float, float]],
) -> typing.List[typing.Tuple[float, float]]:
    """Find the permutation of a sequence of points that minimizes the perimeter
    Euclidean distance.

    Brute force solver for traveling salesman problem

    Parameters
    ----------
    points : List[Tuple[float, float]]
        A list of 2D points where each point is represented as a tuple of (x, y)
        coordinates.

    Returns
    -------
    List[Tuple[float, float]]
        The permutation of points that yields the minimal Euclidean distance.

    Examples
    --------
    >>> points = [(0, 0), (2, 2), (1, 1)]
    >>> minimal_distance_permutation(points)
    [(0, 0), (1, 1), (2, 2)]

    Notes
    -----
    The function uses itertools.permutations to generate all possible orderings
    of the input points, and then calculates the total Euclidean distance for
    each permutation. The permutation with the minimal distance is returned.
    """

    def calc_perimeter_distance(
        perm: typing.Iterable[typing.Tuple[float, float]],
    ) -> float:
        return sum(
            np.linalg.norm(np.array(perm[i - 1]) - np.array(perm[i]))
            for i in range(len(perm))
        )

    min_perm = min(it.permutations(points), key=calc_perimeter_distance)
    return [*min_perm]



from collections import abc
import copy
import typing
import warnings

import frozendict
from matplotlib import axes as mpl_axes
from matplotlib import patches as mpl_patches
from matplotlib import pyplot as plt
import opytional as opyt
import pandas as pd
import seaborn as sns

from ._auxlib.calc_aspect_ import calc_aspect
from ._auxlib.equalize_aspect_ import equalize_aspect
from ._auxlib.set_aspect_ import set_aspect
from ._marqueeplot import marqueeplot, _prepad_axlim
from .mark._MarkMagnifyingGlass import MarkMagnifyingGlass
from .mark._MarkNumericalBadges import MarkNumericalBadges
from .util._NamedFrames import NamedFrames
from .util._SplitKwarg import SplitKwarg


class OutsetGrid(sns.axisgrid.FacetGrid):
    """Facilitates co-display of zoomed-in axis regions transplanted across a
    subplot grid.

    Corresponding regions in the original ("source") plot and the zoomed-in,
    "outset" regions are marked with corresponding "marquee" annotations. The
    OutsetGrid may be configured to include the source plot on the first
    subplot axis, or to only display outset regions on subplot axes.

    Marquee annotation creation must be dispatched manually through
    `marqueeplot`, `marqueeplot_outset`, and/or `marqueeplot_source`. This
    mechanism allows for end-user control over plot sequencing (i.e., layering
    order) and for adjustment of axis sizing prior to marquee rendering.
    (Marquee layout dimensions are sensitive to axis rescaling.)

    Inherits from seaborn's FacetGrid, so FacetGrad API components are
    available, see <https://seaborn.pydata.org/generated/seaborn.FacetGrid.html>.

    Attributes
    ----------
    source_axes : Optional[mpl_axes.Axes]
        The axes object for the source plot, if present.
    outset_axes : Sequence[mpl_axes.Axes]
        The axes objects for the outset plots.

    See Also
    --------
    outset.draw_marquee
        Low-level function for drawing marquee annotations.
    outset.marqueeplot
        Axes-level tidy data interface for creating marquee annotations.
    """

    __data: pd.DataFrame
    _marqueeplot_outset: typing.Callable
    _marqueeplot_source: typing.Callable

    source_axes: typing.Optional[mpl_axes.Axes]
    outset_axes: typing.Sequence[mpl_axes.Axes]

    def add_legend(self: "OutsetGrid", *args, **kwargs) -> None:
        for ax in self.axes.flat:
            ax.legend().set_visible(False)
        if not self._legend_data and self._colors and self.hue_names:
            super().add_legend(
                title=self._hue_var,
                handles=[
                    mpl_patches.Patch(color=c, label=l)
                    for c, l in zip(self._colors, self.hue_names)
                ],
                labels=self.hue_names,
            )
        else:
            super().add_legend(*args, **kwargs)

    def tight_layout(self: "OutsetGrid") -> None:
        self.figure.tight_layout()

    def _finalize_grid(
        self: "OutsetGrid", axlabels: typing.Sequence[str]
    ) -> None:
        """Finalize the annotations and layout."""
        if not self._is_inset():
            # calls to tight layout sometimes cause unwanted reversions
            # of axes insetting --- not sure why
            self.tight_layout()

    def _is_inset(self: "OutsetGrid") -> bool:
        """Are outset axes inset over source axes?"""
        return (
            self.source_axes is not None
            and len(self.outset_axes)
            and self.outset_axes[0].get_position().x0
            < self.source_axes.get_position().corners()[-1][0]
        )

    def __init__(
        self: "OutsetGrid",
        data: typing.Union[
            pd.DataFrame,
            typing.Sequence[typing.Tuple[float, float, float, float]],
            NamedFrames,
        ],
        *,
        x: typing.Optional[str] = None,
        y: typing.Optional[str] = None,
        col: typing.Union[str, bool, None] = None,
        col_order: typing.Optional[typing.Sequence[str]] = None,
        col_wrap: typing.Optional[int] = None,
        hue: typing.Union[str, bool, None] = None,
        hue_order: typing.Optional[typing.Sequence[str]] = None,
        color: typing.Optional[str] = None,  # pass to override outset hues
        include_sourceplot: bool = True,
        marqueeplot_kws: typing.Dict = frozendict.frozendict(),
        marqueeplot_outset_kws: typing.Dict = frozendict.frozendict(),
        marqueeplot_source_kws: typing.Dict = frozendict.frozendict(),
        palette: typing.Optional[typing.Sequence] = None,
        zorder: float = 0.0,
        **kwargs,
    ) -> None:
        """Create an OutsetGrid with specified configuration.

        The arguments `data`, `x`, `y`, `col`, `hue`, and `outset` follow
        seaborn-like tidy data API convention. Note that the same value may
        be specified for more than one of `col`, `hue, and `outset`. Marquee
        annotations are created to contain each subset of x, y values with
        identical `col`, `hue`, and `outset` values.

        Marquee frame coordinates may also be specified directly as a sequence
        of four-element tuples as "extents" `(x0, y0, x1, y1)` or  "boundary
        points" `((x0, y0), (x1, y1))`. In this case, each frame will get its
        own hue and outset plot. For more elaborate manually-specified frame
        layouts, prepare a DataFrame with columns for `x`, `y`, `hue`, and `col`
        with two rows per frame (i.e., one row per boundary point) and pass
        the `frame_inner_pad=0` as a marqueeplot kwarg.

        Marquee annotation creation must be dispatched manually after
        initialization through `marqueeplot`, `marqueeplot_outset`, and/or
        `marqueeplot_source`.

        Parameters
        ----------
        data : pd.DataFrame or Sequence of Tuple[float, float, float, float] or
        outset.util.NamedFrames
            A DataFrame containing the data for plotting, or as a sequence of
            "extents" `(x0, y0, x1, y1)` or  "boundary points" `((x0, y0), (x1,
            y1))` specifying the bounds of outset frames.

            If NamedFrames, underlying data should map frame names to frame
            coordinates.
        x : Optional[str], default None
            Column name to be used for x-axis values.

            If `data` specifies outset frames directly, this kwarg is not
            required. If provided in this case, it will be used as an axis
            label.
        y : Optional[str], default None
            Column name to be used for y-axis values.

            If `data` specifies outset frames directly, this kwarg is not
            required. If provided in this case, it will be used as an axis
            label.
        col : Union[str, bool, None], default None
            Column name for categorical variable to facet across subaxes.

            If None or True, set to match `hue` if provided. If False, no faceting is performed.

            If `data` specifies outset frames directly, this kwarg is not
            required. If provided in this case, it will be used to title
            subplots label.
        col_order : Optional[Sequence[str]], default None
            The order to arrange the columns in the grid.
        col_wrap : Optional[int], default None
            Number of columns where axes grid should wrap to a new row.
        hue : Optional[str], default None
            Column name for categorical variable to determine rendered color of
            data's rendered marquee annotations.

            Will also facet across subaxes if `col` is not set `False`.

            If `data` specifies outset frames directly, this kwarg is not
            required. If provided in this case, it will be used to title legend
            created by `.add_legend()`, if any.
        hue_order : Optional[Sequence[str]], default None
            The to assign palette colors to `hue` categorical values.

            May contain all or a subset of `data[hue]` values.
        color : Optional[str], default None
            Color for all outset annotations. Overrides the palette.
        include_sourceplot : bool, default True
            Whether to include the original source plot in the grid.
        marqueeplot_kws : Dict, default frozendict()
            Keyword arguments to adjust marquee placement and styling over all
            plots.

            See `outset.marqueeplot` for available options.
        marqueeplot_outset_kws : Dict, default frozendict()
            Keyword arguments to adjust marquee placement and styling over
            outset plots.

            See `outset.marqueeplot` for available options.
        marqueeplot_source_kws : Dict, default frozendict()
            Keyword arguments to adjust marquee placement and styling over
            source plot, if present.

            See `outset.marqueeplot` for available options.
        palette : Optional[Sequence], default None
            Color palette for the outset hue sequence.
        zorder : float, default 0.0
            The z-order for plotting elements.
        **kwargs : dict
            Additional keyword arguments forward to seaborn's `FacetGrid`.

            See <https://seaborn.pydata.org/generated/seaborn.FacetGrid.html>
            for details.
        """

        if col is None and col_order is not None:
            raise ValueError("col_order must be None if col not specified")
        if hue is None and hue_order is not None:
            raise ValueError("hue_order must be None if hue not specified")

        for a in "frame_inner_pad":
            if a in marqueeplot_outset_kws or a in marqueeplot_source_kws:
                warnings.warn(
                    f"Specifying {a} independently for only source or "
                    "outset may cause discrepancies in frame placement",
                )

        default_frame_outer_pad_outset = 0.1
        default_frame_outer_pad_source = 0.1
        default_frame_inner_pad = 0.1

        # spoof data frame if outset frames are specified directly
        if isinstance(data, (pd.DataFrame, abc.Mapping)) and not isinstance(
            data, NamedFrames
        ):
            if x is None or y is None:
                raise ValueError(
                    "x and y kwargs must be provided from column names in data",
                )
            if col is None:
                col = hue
                if (
                    col_order is None
                    and hue is not None
                    and hue_order is not None
                ):
                    col_order = hue_order

        else:
            if col is None:
                col = True
            if hue is None and color is None:
                hue = True

            if hue and isinstance(data, NamedFrames):
                hue_order = data.keys()
            if col and isinstance(data, NamedFrames):
                col_order = data.keys()

            x, y = opyt.or_value(x, "_x"), opyt.or_value(y, "_y")
            if col is True:
                col = hue if isinstance(hue, str) else "_outset"
            if hue is True:
                hue = col
            data = pd.DataFrame.from_records(
                [
                    {
                        x: x_,
                        y: y_,
                        col: i,
                        hue: i,
                    }
                    for i, boundary_points in (
                        data.items()
                        if isinstance(data, NamedFrames)
                        else enumerate(data, start=1)
                    )
                    for x_, y_ in (
                        boundary_points
                        if len(boundary_points) == 2
                        else (boundary_points[:2], boundary_points[2:])
                    )
                ],
            )
            default_frame_inner_pad = 0
            default_frame_outer_pad_source = 0

            if len(data) == 0:
                data = pd.DataFrame({x: [], y: [], col: [], hue: []})

        if not x in data.columns:
            raise ValueError(f"kwarg x={x} must be provided as column in data")
        if not y in data.columns:
            raise ValueError(f"kwarg x={y} must be provided as column in data")

        if col is True:
            if hue is None:
                raise ValueError("hue must be provided if col is True")
            col = hue
        elif col is False:
            col = None

        if hue is True:
            if col is None:
                raise ValueError("col must be provided if hue is True")
            hue = col
        elif hue is False:
            hue = None

        if col is not None and not col in data.columns:
            raise ValueError("if provided, col must be a column in data")

        if hue is not None and not hue in data.columns:
            raise ValueError("if provided, hue must be a column in data")

        self.__data = data

        if hue is not None and hue_order is None:
            hue_order = sorted(data[hue].unique())

        if col is None:
            assert "_dummy_col" not in data.columns
            col = "_dummy_col"
            data[col] = 0

        if col_order is None:
            col_order = sorted(data[col].unique())

        if include_sourceplot:
            if None in data[col].unique():
                raise ValueError(
                    "Cannot include source plot if col data includes None",
                )
            col_order = [None] + [*col_order]

        # initialize axes
        #######################################################################
        super().__init__(  # initialize parent FacetGrid
            data,
            col=col,
            col_order=col_order,
            col_wrap=col_wrap,
            hue=hue,
            hue_order=hue_order,
            palette=palette,
            **{
                "legend_out": True,
                "sharex": False,
                "sharey": False,
                **kwargs,
            },
        )

        if col in ("_dummy_col", "_outset"):
            self.set_titles(col_template="")

        if "_x" in data.columns:
            self.set_axis_labels(x_var="")
        else:
            self.set_axis_labels(x_var=x)

        if "_y" in data.columns:
            self.set_axis_labels(y_var="")
        else:
            self.set_axis_labels(y_var=y)

        if include_sourceplot:
            self.source_axes = self.axes.flat[0]
            self.source_axes.set_title("")
            self.outset_axes = self.axes.flat[1:]
        else:
            self.source_axes = None
            self.outset_axes = self.axes.flat[:]

        # draw sourceplot
        #######################################################################
        default_draw_glyph_functor_class = (
            MarkMagnifyingGlass
            if len(self.outset_axes) == 1
            else MarkNumericalBadges
        )

        def marqueeplot_source(self_: "OutsetGrid") -> None:
            if self_.source_axes is None:
                return
            marqueeplot(
                data,
                x=x,
                y=y,
                hue=hue,
                hue_order=hue_order,
                outset=col,
                outset_order=col_order,
                ax=self_.source_axes,
                **{
                    "color": color,
                    "palette": palette,
                    "frame_inner_pad": default_frame_inner_pad,
                    "frame_outer_pad": default_frame_outer_pad_source,
                    "mark_glyph": default_draw_glyph_functor_class,
                    "tight_axlim": False,
                    "zorder": zorder,
                    **copy.deepcopy(marqueeplot_kws),  # for mark_glyph
                    **marqueeplot_source_kws,
                    "frame_edge_kws": {
                        **marqueeplot_kws.get("frame_edge_kws", {}),
                        **marqueeplot_source_kws.get("frame_edge_kws", {}),
                    },
                    "frame_face_kws": {
                        **marqueeplot_kws.get("frame_face_kws", {}),
                        **marqueeplot_source_kws.get("frame_face_kws", {}),
                    },
                    "leader_edge_kws": {
                        **marqueeplot_kws.get("leader_edge_kws", {}),
                        **marqueeplot_source_kws.get("leader_edge_kws", {}),
                    },
                    "leader_face_kws": {
                        **marqueeplot_kws.get("leader_face_kws", {}),
                        **marqueeplot_source_kws.get("leader_face_kws", {}),
                    },
                    "mark_glyph_kws": {
                        "markersize": 16,
                        "zorder": zorder + 1.01,
                        **marqueeplot_kws.get("mark_glyph_kws", {}),
                        **marqueeplot_source_kws.get("mark_glyph_kws", {}),
                    },
                },
            )

        self._marqueeplot_source = marqueeplot_source

        # draw outplots
        #######################################################################
        def marqueeplot_outset(self_: "OutsetGrid") -> None:
            # setup
            for d in marqueeplot_kws, marqueeplot_outset_kws:
                for k, v in d.items():
                    if k == "mark_glyph" and isinstance(v, type):
                        d[k] = v()

            needs_prepad = not (hue is None or hue == col)
            if needs_prepad:
                # need to prepad without split by hue
                prepad_kws = {
                    "frame_inner_pad": default_frame_inner_pad,
                    "frame_outer_pad": default_frame_outer_pad_outset,
                    "frame_outer_pad_unit": "axes",
                }
                fil = data[col].isin(col_order)
                if hue is not None:
                    assert hue_order is not None
                    fil &= data[hue].isin(hue_order)
                self.broadcast_outset(
                    _prepad_axlim,
                    data=data[fil],
                    x=x,
                    y=y,
                    hue=hue,
                    outset=col,
                    tight_axlim=True,
                    **{
                        **prepad_kws,
                        **{
                            k: v
                            for k, v in marqueeplot_kws.items()
                            if k in prepad_kws
                        },
                        **{
                            k: v
                            for k, v in marqueeplot_outset_kws.items()
                            if k in prepad_kws
                        },
                    },
                )

            self_.map_dataframe_outset(
                marqueeplot,
                x=x,
                y=y,
                **{
                    "color": color,
                    "palette": palette,
                    "frame_inner_pad": default_frame_inner_pad,
                    "frame_outer_pad": default_frame_outer_pad_outset,
                    "frame_outer_pad_unit": "axes",
                    "leader_stretch": 0.2,
                    "leader_stretch_unit": "inchesfrom",
                    "mark_glyph": default_draw_glyph_functor_class(),
                    "tight_axlim": not needs_prepad,
                    "zorder": zorder,
                    **marqueeplot_kws,
                    **marqueeplot_outset_kws,
                    "frame_edge_kws": {
                        **marqueeplot_kws.get("frame_edge_kws", {}),
                        **marqueeplot_outset_kws.get("frame_edge_kws", {}),
                    },
                    "frame_face_kws": {
                        **marqueeplot_kws.get("frame_face_kws", {}),
                        **marqueeplot_outset_kws.get("frame_face_kws", {}),
                    },
                    "leader_edge_kws": {
                        **marqueeplot_kws.get("leader_edge_kws", {}),
                        **marqueeplot_outset_kws.get("leader_edge_kws", {}),
                    },
                    "leader_face_kws": {
                        **marqueeplot_kws.get("leader_face_kws", {}),
                        **marqueeplot_outset_kws.get("leader_face_kws", {}),
                    },
                    "mark_glyph_kws": {
                        **({"markersize": 16} if self_._is_inset() else {}),
                        "zorder": zorder + 1.01,
                        **marqueeplot_kws.get("mark_glyph_kws", {}),
                        **marqueeplot_outset_kws.get("mark_glyph_kws", {}),
                    },
                },
            )

        self._marqueeplot_outset = marqueeplot_outset

    def equalize_aspect(self: "OutsetGrid") -> "OutsetGrid":
        """Adjust axes {x,y}lims to ensure an equal xlim-to-ylim ratio across
        all axes.

        Returns
        -------
        OutsetGrid
            Returns self.
        """
        if self.source_axes is not None:
            aspect = calc_aspect(self.source_axes)
            for ax in self.outset_axes:
                set_aspect(ax, aspect)
        else:
            equalize_aspect(self.axes.flat)
        return self

    def marqueeplot(
        self: "OutsetGrid",
        equalize_aspect: bool = True,
        preserve_aspect: bool = False,
    ) -> "OutsetGrid":
        """Dispatch marquee annotation rendering for all subplots --- outset as
        well as source, if included.

        Parameters
        ----------
        equalize_aspect : bool, optional, default: True
            If True, adjusts axes limits to enforce equal ylim height to xlim
            width ratio.
        preserve_aspect : bool, optional, default: False
            If True, restore initial aspect ratio after plotting..

        Returns
        -------
        OutsetGrid
            Returns self.
        """
        self.marqueeplot_source(
            equalize_aspect=False, preserve_aspect=preserve_aspect
        )
        self.marqueeplot_outset(
            equalize_aspect=False, preserve_aspect=preserve_aspect
        )
        if equalize_aspect:
            self.equalize_aspect()
        return self

    def marqueeplot_outset(
        self: "OutsetGrid",
        *,
        equalize_aspect: bool = True,
        preserve_aspect: bool = False,
    ) -> "OutsetGrid":
        """Dispatch marquee annotation rendering for outset plots only

        Parameters
        ----------
        equalize_aspect : bool, optional, default: True
            If True, adjusts axes limits to enforce equal ylim height to xlim
            width ratio.
        preserve_aspect : bool, optional, default: False
            If True, restore initial aspect ratio after plotting..

        Returns
        -------
        OutsetGrid
            Returns self.
        """
        if preserve_aspect and equalize_aspect:
            raise ValueError(
                "may only specify one of {preserve,equalize}_aspect",
            )

        aspects = [calc_aspect(ax) for ax in self.outset_axes]
        self._marqueeplot_outset(self)
        if preserve_aspect:
            for ax, aspect in zip(self.outset_axes, aspects):
                set_aspect(ax, aspect)
        self._marqueeplot_outset = lambda self_: warnings.warn(
            "Redundant call to marqueeplot_outset, marquees were already drawn",
        )
        if equalize_aspect:
            self.equalize_aspect()
        return self

    def marqueeplot_source(
        self: "OutsetGrid",
        *,
        equalize_aspect: bool = True,
        preserve_aspect: bool = False,
    ) -> "OutsetGrid":
        """Dispatch marquee annotation rendering for the source plot only, if
        included.

        Parameters
        ----------
        equalize_aspect : bool, optional, default: True
            If True, adjusts axes limits to enforce equal ylim height to xlim
            width ratio.
        preserve_aspect : bool, optional, default: False
            If True, restore initial aspect ratio after plotting..

        Returns
        -------
        OutsetGrid
            Returns self.
        """
        if preserve_aspect and equalize_aspect:
            raise ValueError(
                "may only specify one of {preserve,equalize}_aspect",
            )

        if self.source_axes is not None:
            aspect = calc_aspect(self.source_axes)
        self._marqueeplot_source(self)
        if preserve_aspect and self.source_axes is not None:
            set_aspect(self.source_axes, aspect)
        self._marqueeplot_source = lambda self_: warnings.warn(
            "Redundant call to marqueeplot_source, marquees were already drawn",
        )
        if equalize_aspect:
            self.equalize_aspect()
        return self

    def map(self: "OutsetGrid") -> None:
        """Placeholder, raises NotImplementedError."""
        raise NotImplementedError()

    def map_outset(self: "OutsetGrid") -> None:
        """Placeholder, raises NotImplementedError."""
        raise NotImplementedError()

    def map_source(self: "OutsetGrid") -> None:
        """Placeholder, raises NotImplementedError."""
        raise NotImplementedError()

    def map_dataframe(
        self: "OutsetGrid", plotter: typing.Callable, *args, **kwargs
    ) -> "OutsetGrid":
        """Map a plotting function over all axes, including source plot axes
        (if present).

        Uses data stored at initialization, passed as a first `data=` kwarg. To
        plot other data, use `broadcast`. Complete dataset is used for source
        axes and faceted subsets are used for each outset axes.

        Parameters
        ----------
        plotter : Callable
            The plotting function to be applied to each axis.
        *args : tuple
            Positional arguments passed to the plotting function.
        **kwargs : dict
            Keyword arguments passed to the plotting function.

        Returns
        -------
        OutsetGrid
            Returns self.
        """
        self.map_dataframe_outset(
            plotter,
            *[
                arg.outset if isinstance(arg, SplitKwarg) else arg
                for arg in args
            ],
            **{
                k: v.outset if isinstance(v, SplitKwarg) else v
                for k, v in kwargs.items()
            },
        )
        self.map_dataframe_source(
            plotter,
            *[
                arg.source if isinstance(arg, SplitKwarg) else arg
                for arg in args
            ],
            **{
                k: v.source if isinstance(v, SplitKwarg) else v
                for k, v in kwargs.items()
            },
        )
        return self

    def map_dataframe_outset(
        self: "OutsetGrid", plotter: typing.Callable, *args, **kwargs
    ) -> "OutsetGrid":
        """Map a plotting function over outset axes only.

        Uses data stored at initialization, passed as a first `data=` kwarg and
        faceted per subplot axes. To plot other data, use `broadcast_outset`.

        Parameters
        ----------
        plotter : Callable
            The plotting function to be applied to each axis.
        *args : tuple
            Positional arguments passed to the plotting function.
        **kwargs : dict
            Keyword arguments passed to the plotting function.

        Returns
        -------
        OutsetGrid
            Returns self.
        """
        if "hue" in kwargs and self._hue_var is not None:
            raise ValueError("Cannot map `hue` if FacetGrid `hue` is set.")
        elif "hue" in kwargs and kwargs.get("hue_order", None) is None:
            assert self.hue_names is None
            kwargs["hue_order"] = sorted(self.__data[kwargs["hue"]].unique())

        xlabels = [ax.get_xlabel() for ax in self.axes.flat]
        ylabels = [ax.get_ylabel() for ax in self.axes.flat]
        super().map_dataframe(plotter, *args, **kwargs)
        if kwargs.get("x", None) == self._x_var and self._x_var is not None:
            for ax, xlabel in zip(self.axes.flat, xlabels):
                ax.set_xlabel(xlabel)
        if kwargs.get("y", None) == self._y_var and self._y_var is not None:
            for ax, ylabel in zip(self.axes.flat, ylabels):
                ax.set_ylabel(ylabel)
        return self

    def map_dataframe_source(
        self: "OutsetGrid", plotter: typing.Callable, *args, **kwargs
    ) -> "OutsetGrid":
        """Map a plotting function over the source plot axes only.

        If source plot axes are not enabled, performs no-op.

        Uses data stored at initialization, passed as a first `data=` kwarg.
        To plot other data, use `broadcast_source`.

        Parameters
        ----------
        plotter : Callable
            The plotting function to be applied to each axis.
        *args : tuple
            Positional arguments passed to the plotting function.
        **kwargs : dict
            Keyword arguments passed to the plotting function.

        Returns
        -------
        OutsetGrid
            Returns self.
        """
        if self._hue_var is not None and "hue" in kwargs:
            raise ValueError("Cannot map `hue` if FacetGrid `hue` is set.")
        if "hue_order" in kwargs and not (
            "hue" in kwargs or self._hue_var is not None
        ):
            raise ValueError("Cannot map `hue_order` if `hue` is unset.")
        elif "hue_order" in kwargs and self.hue_names is None:
            raise ValueError(
                "Cannot map `hue_order` if FacetGrid `hue_order` is set.",
            )

        hue = opyt.or_value(self._hue_var, kwargs.pop("hue", None))
        hue_order = opyt.or_value(self.hue_names, kwargs.pop("hue_order", None))
        if hue is not None:
            kwargs["hue"] = hue
        if hue_order is not None:
            kwargs["hue_order"] = hue_order
        if self.source_axes is not None:
            xlabel, ylabel = (
                self.source_axes.get_xlabel(),
                self.source_axes.get_ylabel(),
            )
            plotter(self.__data, *args, ax=self.source_axes, **kwargs)
            if kwargs.get("x", None) == self._x_var and self._x_var is not None:
                self.source_axes.set_xlabel(xlabel)
            if kwargs.get("y", None) == self._y_var and self._y_var is not None:
                self.source_axes.set_ylabel(ylabel)
        self.tight_layout()
        return self

    def broadcast(
        self: "OutsetGrid",
        plotter: typing.Callable,
        *args,
        **kwargs,
    ) -> "OutsetGrid":
        """Map a plotting function over all axes, including the source plot
        axis (if present).

        Performs call with same data and arguments for all axes. To use faceted
        data stored at initialization, refer to `map_dataframe`.

        Parameters
        ----------
        plotter : Callable
            The plotting function to be applied to each axis.
        *args : tuple
            Positional arguments passed to the plotting function.
        **kwargs : dict
            Keyword arguments passed to the plotting function.

        Returns
        -------
        OutsetGrid
            Returns self.

        Notes
        -----
        Does not use data stored from initialization. Data should be provided
        via argument to this method.

        Preserves axis limits for all axes except the source plot, if present.
        """
        self.broadcast_outset(
            plotter,
            *[
                arg.outset if isinstance(arg, SplitKwarg) else arg
                for arg in args
            ],
            **{
                k: v.outset if isinstance(v, SplitKwarg) else v
                for k, v in kwargs.items()
            },
        )
        self.broadcast_source(
            plotter,
            *[
                arg.source if isinstance(arg, SplitKwarg) else arg
                for arg in args
            ],
            **{
                k: v.source if isinstance(v, SplitKwarg) else v
                for k, v in kwargs.items()
            },
        )
        return self

    def broadcast_outset(
        self: "OutsetGrid",
        plotter: typing.Callable,
        *args,
        **kwargs,
    ) -> "OutsetGrid":
        """Map a plotting function over only outset axes.

        Performs call with same data and arguments for all axes. To use faceted
        data stored at initialization, refer to `map_dataframe`.

        Parameters
        ----------
        plotter : Callable
            The plotting function to be applied to each axis.
        *args : tuple
            Positional arguments passed to the plotting function.
        **kwargs : dict
            Keyword arguments passed to the plotting function.

        Returns
        -------
        OutsetGrid
            Returns self.

        Notes
        -----
        Does not use data stored from initialization. Data should be provided
        via argument to this method.

        Preserves axis limits.
        """
        xlabels = [ax.get_xlabel() for ax in self.axes.flat]
        ylabels = [ax.get_ylabel() for ax in self.axes.flat]
        for ax in self.outset_axes:
            # store and restore axis limits, except for source plot if present
            xlim, ylim = ax.get_xlim(), ax.get_ylim()
            try:
                plotter(*args, ax=ax, **kwargs)
            except (TypeError, AttributeError):
                plt.sca(ax)
                plotter(*args, **kwargs)
            ax.set_xlim(*xlim)
            ax.set_ylim(*ylim)
        if kwargs.get("x", None) == self._x_var and self._x_var is not None:
            for ax, xlabel in zip(self.axes.flat, xlabels):
                ax.set_xlabel(xlabel)
        if kwargs.get("y", None) == self._y_var and self._y_var is not None:
            for ax, ylabel in zip(self.axes.flat, ylabels):
                ax.set_ylabel(ylabel)
        self.tight_layout()
        return self

    def broadcast_source(
        self: "OutsetGrid",
        plotter: typing.Callable,
        *args,
        **kwargs,
    ) -> "OutsetGrid":
        """Map a plotting function over source plot axes, if present.

        Performs call with same data and arguments for all axes. To use faceted
        data stored at initialization, refer to `map_dataframe`.

        Parameters
        ----------
        plotter : Callable
            The plotting function to be applied to each axis.
        *args : tuple
            Positional arguments passed to the plotting function.
        **kwargs : dict
            Keyword arguments passed to the plotting function.

        Returns
        -------
        OutsetGrid
            Returns self.

        Notes
        -----
        Does not use data stored from initialization. Data should be provided
        via argument to this method.

        Doesn't preserve axis limits.
        """
        if self.source_axes is None:
            return self
        xlabel, ylabel = (
            self.source_axes.get_xlabel(),
            self.source_axes.get_ylabel(),
        )
        ax = self.source_axes
        try:
            plotter(*args, ax=ax, **kwargs)
        except (TypeError, AttributeError):
            plt.sca(ax)
            plotter(*args, **kwargs)
        self.tight_layout()
        if kwargs.get("x", None) == self._x_var and self._x_var is not None:
            self.source_axes.set_xlabel(xlabel)
        if kwargs.get("y", None) == self._y_var and self._y_var is not None:
            self.source_axes.set_ylabel(ylabel)
        return self



import matplotlib.pyplot as plt
import pytest

from outset._auxlib.calc_outer_pad_ import calc_outer_pad


@pytest.mark.parametrize("pad_unit", ["axes", "figure", "inches"])
def test_calc_outer_pad(pad_unit: str):
    ax = plt.figure().add_subplot(111)
    pad = 0.1
    result = calc_outer_pad(ax, pad, pad_unit)
    assert isinstance(result, tuple)



import matplotlib.pyplot as plt
import numpy as np
import pytest
import seaborn as sns

from outset.stub import rescale_clip_outliers


def test_rescale_clip_outliers():
    np.random.seed(0)
    x = np.append(np.random.normal(0, 1, 100), [10, 15])  # with two outliers
    y = np.append(np.random.normal(0, 1, 100), [10, 20])

    fig, ax = plt.subplots()
    ax.scatter(x, y)

    # Apply the rescale_clip_outliers function
    xlim_, ylim_ = ax.get_xlim(), ax.get_ylim()
    rescale_clip_outliers(ax)
    assert xlim_ != ax.get_xlim() and ylim_ != ax.get_ylim()

    # Get new axis limits after rescaling
    x0, x1 = ax.get_xlim()
    y0, y1 = ax.get_ylim()

    # Check if outliers are clipped
    assert len(x) - 10 <= sum(x0 <= x_ <= x1 for x_ in x) < len(x)
    assert len(y) - 10 <= sum(y0 <= y_ <= y1 for y_ in y) < len(y)


def test_rescale_clip_outliers_with_seaborn():
    np.random.seed(0)
    x = np.append(np.random.normal(0, 1, 100), [10, 15])  # with two outliers
    y = np.append(np.random.normal(0, 1, 100), [10, 20])
    data = {"x": x, "y": y}

    # Create a Seaborn plot
    ax = sns.scatterplot(x="x", y="y", data=data)

    # Apply the rescale_clip_outliers function
    xlim_, ylim_ = ax.get_xlim(), ax.get_ylim()
    rescale_clip_outliers(ax)
    assert xlim_ != ax.get_xlim() and ylim_ != ax.get_ylim()

    # Get new axis limits after rescaling
    x0, x1 = ax.get_xlim()
    y0, y1 = ax.get_ylim()

    # Check if outliers are clipped
    assert len(x) - 10 <= sum(x0 <= x_ <= x1 for x_ in x) < len(x)
    assert len(y) - 10 <= sum(y0 <= y_ <= y1 for y_ in y) < len(y)


def test_rescale_clip_empty():
    fig, ax = plt.subplots()
    ax.scatter([], [])

    xlim_, ylim_ = ax.get_xlim(), ax.get_ylim()
    rescale_clip_outliers(ax)

    assert xlim_ == ax.get_xlim()
    assert ylim_ == ax.get_ylim()


@pytest.mark.parametrize("pad", [0.01, 0.1])
def test_rescale_clip_singleton(pad: float):
    fig, ax = plt.subplots()
    ax.scatter([3], [4])

    rescale_clip_outliers(
        calc_outlier_bounds=lambda x: (min(x), max(x)),
        pad=pad,
    )

    assert (3 - 3 * pad, 3 + 3 * pad) == ax.get_xlim()
    assert (4 - 4 * pad, 4 + 4 * pad) == ax.get_ylim()



import typing

from matplotlib import markers as mpl_markers
from matplotlib.path import Path as mpl_Path


# adapted from https://stackoverflow.com/a/26726237
def align_marker(
    marker: str,
    halign: typing.Literal["left", "center", "right"] = "center",
    valign: typing.Literal["top", "middle", "bottom"] = "middle",
    pad: float = 1.0,
) -> mpl_Path:
    """Create markers with specified alignment.

    Parameters
    ----------
    marker : a valid marker specification.
      See mpl.markers
    halign : string, float {'left', 'center', 'right'}
      Specifies the horizontal alignment of the marker. *float* values
      specify the alignment in units of the markersize/2 (0 is 'center',
      -1 is 'right', 1 is 'left').
    valign : string, float {'top', 'middle', 'bottom'}
      Specifies the vertical alignment of the marker. *float* values
      specify the alignment in units of the markersize/2 (0 is 'middle',
      -1 is 'top', 1 is 'bottom').

    Returns
    -------
    marker_array : numpy.ndarray
      A Nx2 array that specifies the marker path relative to the
      plot target point at (0, 0).

    Notes
    -----
    The mark_array can be passed directly to ax.plot and ax.scatter, e.g.::

        ax.plot(1, 1, marker=align_marker('>', 'left'))
    """

    if isinstance(halign, str):
        halign = {
            "right": -pad,
            "middle": 0.0,
            "center": 0.0,
            "left": pad,
        }[halign]

    if isinstance(valign, str):
        valign = {
            "top": -pad,
            "middle": 0.0,
            "center": 0.0,
            "bottom": pad,
        }[valign]

    # Define the base marker
    bm = mpl_markers.MarkerStyle(marker)

    # Get the marker path and apply the marker transform to get the
    # actual marker vertices (they should all be in a unit-square
    # centered at (0, 0))
    m_arr = bm.get_path().transformed(bm.get_transform()).vertices

    # Shift the marker vertices for the specified alignment.
    m_arr[:, 0] += halign / 2
    m_arr[:, 1] += valign / 2

    return mpl_Path(m_arr, bm.get_path().codes)
